@@ -114,6 +114,7 @@ int Create(struct supertype *st, char *mddev,
 	struct mdinfo info, *infos;
 	int did_default = 0;
 	int do_default_layout = 0;
+	int do_default_chunk = 0;
 	unsigned long safe_mode_delay = 0;
 	char chosen_name[1024];
 	struct map_ent *map = NULL;
@@ -229,12 +230,10 @@ int Create(struct supertype *st, char *mddev,
 	case 10:
 	case 6:
 	case 0:
-		if (chunk == 0) {
-			if (st && st->ss->default_geometry)
-				st->ss->default_geometry(st, NULL, NULL, &chunk);
-			chunk = chunk ? : 512;
-			if (verbose > 0)
-				fprintf(stderr, Name ": chunk size defaults to %dK\n", chunk);
+		if (chunk == 0 || chunk == UnSet) {
+			chunk = UnSet;
+			do_default_chunk = 1;
+			/* chunk will be set later */
 		}
 		break;
 	case LEVEL_LINEAR:
@@ -260,12 +259,25 @@ int Create(struct supertype *st, char *mddev,
 		return 1;
 	}
 	
-	if (size && chunk)
+	if (size && chunk && chunk != UnSet)
 		size &= ~(unsigned long long)(chunk - 1);
 	newsize = size * 2;
 	if (st && ! st->ss->validate_geometry(st, level, layout, raiddisks,
-					      chunk, size*2, NULL, &newsize, verbose>=0))
+					      &chunk, size*2, NULL, &newsize, verbose>=0))
 		return 1;
+
+	if (chunk && chunk != UnSet) {
+		newsize &= ~(unsigned long long)(chunk*2 - 1);
+		if (do_default_chunk) {
+			/* default chunk was just set */
+			if (verbose > 0)
+				fprintf(stderr, Name ": chunk size "
+					"defaults to %dK\n", chunk);
+			size &= ~(unsigned long long)(chunk - 1);
+			do_default_chunk = 0;
+		}
+	}
+
 	if (size == 0) {
 		size = newsize / 2;
 		if (size && verbose > 0)
@@ -280,6 +292,8 @@ int Create(struct supertype *st, char *mddev,
 	for (dv=devlist; dv && !have_container; dv=dv->next, dnum++) {
 		char *dname = dv->devname;
 		unsigned long long freesize;
+		int dfd;
+
 		if (strcasecmp(dname, "missing")==0) {
 			if (first_missing > dnum)
 				first_missing = dnum;
@@ -288,6 +302,20 @@ int Create(struct supertype *st, char *mddev,
 			missing_disks ++;
 			continue;
 		}
+		dfd = open(dname, O_RDONLY);
+		if (dfd < 0) {
+			fprintf(stderr, Name ": cannot open %s: %s\n",
+				dname, strerror(errno));
+			exit(2);
+		}
+		if (fstat(dfd, &stb) != 0 ||
+		    (stb.st_mode & S_IFMT) != S_IFBLK) {
+			close(dfd);
+			fprintf(stderr, Name ": %s is not a block device\n",
+				dname);
+			exit(2);
+		}
+		close(dfd);
 		info.array.working_disks++;
 		if (dnum < raiddisks)
 			info.array.active_disks++;
@@ -308,12 +336,21 @@ int Create(struct supertype *st, char *mddev,
 					layout = default_layout(st, level, verbose);
 				if (st && !st->ss->validate_geometry
 					    	(st, level, layout, raiddisks,
-						 chunk, size*2, dname, &freesize,
-						 verbose > 0))
+						 &chunk, size*2, dname, &freesize,
+						 verbose > 0)) {
+					free(st);
 					st = NULL;
+					chunk = do_default_chunk ? UnSet : chunk;
+				}
 			}
 
 			if (!st) {
+				int dfd = open(dname, O_RDONLY|O_EXCL);
+				if (dfd < 0) {
+					fprintf(stderr, Name ": cannot open %s: %s\n",
+						dname, strerror(errno));
+					exit(2);
+				}
 				fprintf(stderr, Name ": device %s not suitable "
 					"for any style of array\n",
 					dname);
@@ -327,7 +364,7 @@ int Create(struct supertype *st, char *mddev,
 				layout = default_layout(st, level, verbose);
 			if (!st->ss->validate_geometry(st, level, layout,
 						       raiddisks,
-						       chunk, size*2, dname,
+						       &chunk, size*2, dname,
 						       &freesize,
 						       verbose >= 0)) {
 
@@ -341,9 +378,17 @@ int Create(struct supertype *st, char *mddev,
 		}
 
 		freesize /= 2; /* convert to K */
-		if (chunk) {
+		if (chunk && chunk != UnSet) {
 			/* round to chunk size */
 			freesize = freesize & ~(chunk-1);
+			if (do_default_chunk) {
+				/* default chunk was just set */
+				if (verbose > 0)
+					fprintf(stderr, Name ": chunk size "
+						"defaults to %dK\n", chunk);
+				size &= ~(unsigned long long)(chunk - 1);
+				do_default_chunk = 0;
+			}
 		}
 
 		if (size && freesize < size) {
@@ -375,13 +420,14 @@ int Create(struct supertype *st, char *mddev,
 			if (strcmp(st->ss->name, "1.x") == 0 &&
 			    st->minor_version >= 1)
 				/* metadata at front */
-				warn |= check_partitions(fd, dname, 0);
-			else if (level == 1 || level == LEVEL_CONTAINER)
+				warn |= check_partitions(fd, dname, 0, 0);
+			else if (level == 1 || level == LEVEL_CONTAINER
+				    || (level == 0 && raiddisks == 1))
 				/* partitions could be meaningful */
-				warn |= check_partitions(fd, dname, freesize*2);
+				warn |= check_partitions(fd, dname, freesize*2, size*2);
 			else
 				/* partitions cannot be meaningful */
-				warn |= check_partitions(fd, dname, 0);
+				warn |= check_partitions(fd, dname, 0, 0);
 			if (strcmp(st->ss->name, "1.x") == 0 &&
 			    st->minor_version >= 1 &&
 			    did_default &&
@@ -420,7 +466,7 @@ int Create(struct supertype *st, char *mddev,
 			/* size is meaningful */
 			if (!st->ss->validate_geometry(st, level, layout,
 						       raiddisks,
-						       chunk, minsize*2,
+						       &chunk, minsize*2,
 						       NULL, NULL, 0)) {
 				fprintf(stderr, Name ": devices too large for RAID level %d\n", level);
 				return 1;
@@ -612,8 +658,8 @@ int Create(struct supertype *st, char *mddev,
 		goto abort;
 
 	total_slots = info.array.nr_disks;
-	sysfs_init(&info, mdfd, 0);
 	st->ss->getinfo_super(st, &info, NULL);
+	sysfs_init(&info, mdfd, 0);
 
 	if (did_default && verbose >= 0) {
 		if (is_subarray(info.text_version)) {
@@ -810,11 +856,19 @@ int Create(struct supertype *st, char *mddev,
 					/* getinfo_super might have lost these ... */
 					inf->disk.major = major(stb.st_rdev);
 					inf->disk.minor = minor(stb.st_rdev);
+					/* FIXME the following should not be needed
+					 * as getinfo_super is suppose to set
+					 * them.  However it doesn't for imsm,
+					 * so we have this hack for now
+					 */
+					if (st->ss == &super_imsm) {
+						inf->disk.number = dnum;
+						inf->disk.raid_disk = dnum;
+					}
 				}
 				break;
 			case 2:
 				inf->errors = 0;
-				rv = 0;
 
 				rv = add_disk(mdfd, st, &info, inf);
 
@@ -850,7 +904,13 @@ int Create(struct supertype *st, char *mddev,
 				me = map_by_devnum(&map, st->container_dev);
 			}
 
-			st->ss->write_init_super(st);
+			if (st->ss->write_init_super(st)) {
+				fprintf(stderr,
+					Name ": Failed to write metadata to %s\n",
+					dv->devname);
+				st->ss->free_super(st);
+				goto abort;
+			}
 
 			/* update parent container uuid */
 			if (me) {
@@ -906,6 +966,10 @@ int Create(struct supertype *st, char *mddev,
 			if (ioctl(mdfd, RUN_ARRAY, &param)) {
 				fprintf(stderr, Name ": RUN_ARRAY failed: %s\n",
 					strerror(errno));
+				if (info.array.chunk_size & (info.array.chunk_size-1)) {
+					fprintf(stderr, "     : Problem may be that "
+						"chunk size is not a power of 2\n");
+				}
 				ioctl(mdfd, STOP_ARRAY, NULL);
 				goto abort;
 			}
@@ -916,7 +980,7 @@ int Create(struct supertype *st, char *mddev,
 			if (need_mdmon)
 				start_mdmon(st->container_dev);
 
-			ping_monitor(devnum2devname(st->container_dev));
+			ping_monitor_by_id(st->container_dev);
 			close(container_fd);
 		}
 		wait_for(chosen_name, mdfd);
